from functools import reduce, wraps
from operator import xor, and_
from typing import List, Dict, Union

import networkx as nx

from weaveio.writequery import CypherVariable, CypherData
from weaveio.writequery.base import BaseStatement, CypherVariableItem

def typeerror_is_false(func):
    @wraps(func)
    def inner(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except TypeError:
            return False
    return inner


def sort_rooted_dag(graph):
    for n, d in graph.in_degree():
        if d == 0:
            break
    return nx.algorithms.traversal.dfs_tree(graph, n)


class Step:
    def __init__(self, direction: str, label: str = 'is_required_by', properties: Dict = None):
        if isinstance(direction, Step):
            self.direction = direction.direction
            self.label = direction.label
            self.properties = direction.properties
        elif direction in ['->', '<-', '-']:
            self.direction = direction
            self.label = label
            self.properties = properties
        else:
            raise ValueError(f"Direction {direction} is not supported")

    @typeerror_is_false
    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        return self.direction == other.direction and self.properties == other.properties and self.label == other.label

    def __str__(self):
        if self.properties is None:
            mid = f'-[:{self.label}]-'
        else:
            mid = f'-[:{self.label} {self.properties}]-'
        if self.direction == '->':
            return f"{mid}"
        elif self.direction == '<-':
            return f"{mid}"
        else:
            return mid


class TraversalPath:
    def __init__(self, *path: Union[Step, str]):
        self._path = path
        self.nodes = []
        self.steps = []
        self.path = []
        self.end = CypherVariable(str(path[-1])) if len(path) > 1 else None
        self.repr_path = ''.join(path)
        for i, entry in enumerate(path[:-1]):
            if not i % 2:  # even number
                step = Step(entry)
                self.steps.append(step)
                self.path.append(step)
            else:
                self.nodes.append(str(entry))
                self.path.append(f'(:{entry})')

    def __len__(self):
        return len(self.nodes) + 1

    def __str__(self):
        if self.end is None:
            return ''
        end = f'({self.end}:{self.end.namehint})'
        return ''.join(map(str, self.path)) + end

    def __repr__(self):
        return f'<TraversalPath({str(self.repr_path)})>'

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        return self._path == other._path

    def __hash__(self):
        return hash(self._path)


class Action(BaseStatement):
    compare = None
    shape = None
    is_filter = False

    def to_cypher(self):
        raise NotImplementedError

    def __getitem__(self, item: CypherVariable):
        if isinstance(item, CypherVariableItem):
            return self.transformed_variables[item.parent].get(item.args)
        return self.transformed_variables[item]

    def __init__(self, input_variables: List[CypherVariable], output_variables: List[CypherVariable],
                 hidden_variables: List[CypherVariable] = None, transformed_variables: Dict[CypherVariable, CypherVariable] = None, target: CypherVariable = None):
        super(Action, self).__init__(input_variables, output_variables, hidden_variables)
        self.transformed_variables = {} if transformed_variables is None else transformed_variables
        self.target = target

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False
        base = set(self.input_variables) == set(other.input_variables) and self.__class__ is other.__class__
        for c in self.compare:
            selfthing = getattr(self, c, None)
            otherthing = getattr(other, c, None)
            base &= selfthing == otherthing
        return base

    def __hash__(self):
        base = reduce(xor, map(hash, [tuple(self.input_variables), self.__class__.__name__]))
        for c in self.compare:
            base ^= hash(getattr(self, c))
        return base

    def __str__(self):
        raise NotImplementedError

    def __repr__(self):
        return f'<{str(self)}>'


class EntryPoint(Action):
    compare = []

    def __init__(self):
        super().__init__([], [])

    def to_cypher(self):
        return ''

    def __str__(self):
        return 'EntryPoint'


class StartingPoint(Action):
    compare = ['labels']

    def __init__(self, *labels):
        self.labels = labels
        self.hierarchy = CypherVariable(self.labels[-1])
        super().__init__([], [self.hierarchy], target=self.hierarchy)

    def to_cypher(self):
        s = f"OPTIONAL MATCH ({self.hierarchy}:{self.labels[0]})"
        if len(self.labels) == 1:
            return s
        condition = ' OR '.join([f'{self.hierarchy}:{l}' for l in self.labels[1:]])
        return s + f" WHERE {condition}"

    def __str__(self):
        return '|'.join(self.labels)


class DataReference(Action):
    compare = ['hashes']

    def __init__(self, *data):
        import numpy as np
        self.hashes = reduce(xor, [hash(np.array(a).tobytes()) for a in data])
        ins = [CypherData(datum, delay=True) for datum in data]
        super().__init__(ins, [])

    def to_cypher(self):
        return '// added data here'

    def __str__(self):
        return 'DataReference'


class Traversal(Action):
    """
    Traverse from one hierarchy level to another. This extends the branch and
    potentially increases the cardinality.
    Given a `source` branch and one or more `paths` of form (minnumber, maxnumber, direction),
     traverse to the nodes described by the `paths`.
    For example:
        >>> Traversal(branch, TraversalPath(['->', 'Exposure', '->', 'OB', '->', 'OBSpec']))
        results in `OPTIONAL MATCH (run)-[]->(:Exposure)-[]->(:OB)-[]->(obspec:OBSpec)`
    To traverse multiple paths at once, we use unions in a subquery
    """
    compare = ['paths', 'source']

    def __init__(self, source: CypherVariable, *paths: TraversalPath, name=None):
        if name is None:
            name = ''.join(getattr(p.end, 'namehint', source.namehint) for p in paths[:2])
        self.out = CypherVariable(name)
        # if there is an empty path, then we just refer to the source node
        self.ends = [p.end if p.end is not None else source for p in paths]
        super(Traversal, self).__init__([source], [self.out], [i for i in self.ends if i is not source], target=self.out)
        self.source = source
        self.paths = paths

    def to_cypher(self):
        lines = [f'OPTIONAL MATCH ({self.source}){p}' for p in self.paths]
        lines = '\n\nUNION\n\n'.join([f'\tWITH {self.source}\n\t{l}\n\tRETURN DISTINCT {end} as {self.out}' for l, end in zip(lines, self.ends)])
        return f"""CALL {{\n{lines}\n}}"""

    def __str__(self):
        return f'{self.source.namehint}->{self.out.namehint}'


class Return(Action):
    compare = ['branch', 'varnames']  # just compare input_variables

    def __init__(self, branch: 'Branch', *varnames):
        self.branch = branch
        self.varnames = varnames
        super(Return, self).__init__([branch.hierarchies[-1]], [])

    def to_cypher(self):
        proj = ', '.join([f'{self.branch.hierarchies[-1].get(v)}' for v in self.varnames])
        return f"RETURN {proj}"

    def __str__(self):
        return f'return {self.varnames}'


class Collection(Action):
    shape = 'rect'
    compare = ['_singles', '_multiples', '_reference']

    def __init__(self, reference: 'Branch', singles: List['Branch'], multiples: List['Branch']):
        self._singles = tuple(singles)
        self._multiples = tuple(multiples)
        self._reference = reference

        self.references = reference.find_hierarchies()
        self.references += [v for v in reference.find_variables() if v not in self.references and not isinstance(v, CypherData)]
        self.insingle_hierarchies = [h for x in singles for h in x.find_hierarchies() if h not in self.references and not isinstance(h, CypherData)]
        self.insingle_variables = [v for x in singles for v in x.find_variables() if v not in self.insingle_hierarchies
                                   and v not in self.references and not isinstance(v, CypherData)]
        self.inmultiple_hierarchies = [h for x in multiples  for h in x.find_hierarchies() if h not in self.references and not isinstance(h, CypherData)]
        self.inmultiple_variables = [v for x in multiples for v in x.variables if v not in self.insingle_hierarchies
                                     and v not in self.references and not isinstance(v, CypherData)]

        self.outsingle_hierarchies = [CypherVariable(s.namehint) for s in self.insingle_hierarchies]
        self.outsingle_variables = [CypherVariable(s.namehint) for s in self.insingle_variables]
        self.outmultiple_hierarchies = [CypherVariable(s.namehint+'_list') for s in self.inmultiple_hierarchies]
        self.outmultiple_variables = [CypherVariable(s.namehint+'_list') for s in self.inmultiple_variables]
        inputs = self.insingle_hierarchies + self.insingle_variables + self.inmultiple_variables + self.inmultiple_hierarchies
        outputs = self.outsingle_hierarchies + self.outsingle_variables + self.outmultiple_variables + self.outmultiple_hierarchies
        super().__init__(inputs + self.references, outputs, [], transformed_variables={i: o for i, o in zip(inputs, outputs)})

    def to_cypher(self):
        base = [f'{r}' for r in self.references + ['time0']]
        single_hierarchies = [f'head(collect({i})) as {o}' for i, o in zip(self.insingle_hierarchies, self.outsingle_hierarchies)]
        multiple_hierarchies = [f'collect({i}) as {o}' for i, o in zip(self.inmultiple_hierarchies, self.outmultiple_hierarchies)]
        single_variables = [f'head(collect({i})) as {o}' for i, o in zip(self.insingle_variables, self.outsingle_variables)]
        multiple_variables = [f'collect({i}) as {o}' for i, o in zip(self.inmultiple_variables, self.outmultiple_variables)]
        return 'WITH ' + ', '.join(base + single_hierarchies + single_variables + multiple_hierarchies + multiple_variables)

    def __str__(self):
        return f'collect'


class Aggregation(Action):
    shape = 'rect'
    compare = ['string_function', 'variable', 'branch', 'reference', 'remove_infs']
    def __init__(self, string_function: str, variable: CypherVariable, branch: 'Branch', reference: 'Branch', remove_infs: bool, namehint: str):
        self.string_function = string_function
        self.variable = variable
        self.branch = branch
        self.reference = reference
        self.remove_infs = remove_infs
        ins = self.reference.find_variables() + [self.variable]
        self.output = CypherVariable(f"{namehint}_{variable.namehint}")
        transformed_variables = {i: i for i in ins[:-1]}
        transformed_variables[self.variable] = self.output
        super().__init__(ins, [self.output], [], transformed_variables, self.output)

    def to_cypher(self):
        if self.remove_infs:
            string_function = self.string_function.format(x=f"CASE WHEN {self.variable} > apoc.math.maxLong() THEN null ELSE {self.variable} END")
        else:
            string_function = self.string_function.format(x=self.variable)
        statics = ", ".join(map(str, ['time0'] + self.input_variables[:-1]))
        return f'WITH {statics}, {string_function} as {self.output}'

    def __str__(self):
        return f"{self.string_function}"


class Operation(Action):
    compare = ['string_functions', 'hashable_inputs']

    def __init__(self, *string_functions: str, namehint=None, **inputs):
        self.string_functions = string_functions
        if not isinstance(namehint, (list, tuple)):
            namehint = [namehint] * len(string_functions)
        self.outputs = [CypherVariable(nh) for _, nh in zip(string_functions, namehint)]
        self.inputs = inputs
        self.hashable_inputs = tuple(self.inputs.items())
        super().__init__(list(inputs.values()), self.outputs, target=self.outputs[0])

    def to_cypher(self):
        assignments = [f"{func.format(**self.inputs)} as {out}" for func, out in zip(self.string_functions, self.outputs)]
        return f"WITH *, {', '.join(assignments)}"

    def __str__(self):
        return ', '.join(self.string_functions)


class Filter(Operation):
    shape = 'diamond'
    is_filter = True

    def __init__(self, string_function, **inputs):
        super().__init__(string_function, **inputs)
        self.string_function = string_function

    def to_cypher(self):
        return f"WHERE {self.string_function.format(**self.inputs)}"


def shared_branch(*branches: 'Branch'):
    """
    Find the ancestral branch of two or more branches
    """
    newbranches = reduce(and_, [set(p.find_hierarchy_branches()) for p in branches])
    distances = [(ancestor, sum(nx.shortest_path_length(branch.relevant_graph, ancestor, branch) for branch in branches)) for ancestor in newbranches]
    try:
        return max(distances, key=lambda x: x[1])[0]
    except ValueError:
        return branches[0].handler.entry  # no shared parents means going back to the beginning


def shared_branch_without_filters(a: 'Branch', b: 'Branch'):
    shared = shared_branch(a, b)
    if shared is a.handler.entry:
        return shared  # cant go back beyond the start
    afilters = any(i.action.is_filter for i in nx.shortest_path(a.accessible_graph, shared, a))
    bfilters = any(i.action.is_filter for i in nx.shortest_path(b.accessible_graph, shared, b))
    if afilters or bfilters:
        try:
            return shared.find_hierarchy_branches(True)[-2]
        except IndexError:
            return a.handler.entry
    return shared


# class Alignment(Action):
#     compare = ['reference', 'branches']
#
#     def __init__(self, reference: 'Branch', *branches):
#         assert isinstance(reference.action, Collection)
#         for branch in branches:
#             assert isinstance(branch.action, Collection)
#         self.reference = reference
#         self.branches = branches
#         self.ins = list(set(i for x in (reference, ) + branches for i in x.find_variables()))
#         self.outs = [CypherVariable(i.namehint) for i in self.ins]
#         self.indexer = CypherVariable('i')
#         transformed_variables = {i: o for i, o in zip(self.ins, self.outs)}
#         super(Alignment, self).__init__(self.ins, self.outs, [self.indexer], transformed_variables)
#
#     def to_cypher(self):
#         # unwind = f'UNWIND range(0, apoc.coll.max([x in {self.ins} | size(x)])-1) as {self.indexer}'
#         # get = [f'{i}[{self.indexer}] as {o}' for i, o in zip(self.ins, self.outs)]
#         # return f"{unwind}\nWITH *, {', '.join(get)}"
#         return ''
#
#     def __str__(self):
#         return 'align'

class Alignment(Action):
    compare = ['reference', 'branches']
    continues_on = False

    def __str__(self):
        return f'{self.__class__.__name__}'


class Slice(Action):
    compare = []
    is_filter = True

    def __init__(self, slc):
        self.slc = slc
        self.skip = slc.start
        self.limit = slc.stop - slc.start
        super().__init__([], [], [])

    def __str__(self):
        return f'{self.slc}'

    def to_cypher(self):
        return f'WITH * SKIP {self.skip} LIMIT {self.limit}'


class Results(Action):
    compare = ['branches']

    def __init__(self, branch_attributes):
        self.branch_attributes = branch_attributes
        self.branches = tuple(branch_attributes.keys())
        ins = [j for i in self.branch_attributes.values() for j in i]
        super(Results, self).__init__(ins, [], [], {}, None)

    def to_cypher(self):
        return 'RETURN {}'.format(', '.join(map(str, self.input_variables)))

    def __str__(self):
        names = [i.namehint for i in self.input_variables]
        return 'return {}'.format(', '.join(names))


class DifferentLevelAlignment(Alignment):
    compare = ['reference', 'branches']
    continues_on = False

    def __init__(self, reference, branches):
        self.reference = reference
        self.branches = tuple(branches)
        self.shared_variables = reference.find_variables()
        self.to_unwind = [v for branch in branches for v in branch.find_variables() if v not in self.shared_variables]
        ins = self.shared_variables + self.to_unwind
        self.outs = [CypherVariable(v.namehint) for v in self.to_unwind]
        self.indexer = CypherVariable('i')
        transformed = {v: o for v, o in zip(self.to_unwind, self.outs)}
        transformed.update({v: v for v in self.shared_variables})
        super().__init__(ins, self.outs, [self.indexer], transformed)

    def to_cypher(self):
        unwind = f'UNWIND range(0, size({self.to_unwind[0]})-1) as {self.indexer}'
        unzip = [f'{v}[{self.indexer}] as {o}' for v, o in zip(self.to_unwind, self.outs)]
        static = [f'{v}' for v in self.shared_variables]
        get = f'WITH {", ".join(["time0"] + static + unzip)}'
        return f'{unwind}\n{get}'


class ScalarAlignment(Alignment):
    continues_on = True

    def __init__(self, reference, branches):
        self.reference = reference
        self.branches = tuple(branches)
        self.outs = []
        super().__init__([], [])

    def to_cypher(self):
        return f""



from datetime import datetime
from time import sleep
from warnings import warn

import py2neo
from astropy.table import Table
from py2neo import Graph as NeoGraph

import numpy as np
import pandas as pd
from weaveio.context import ContextMeta
from weaveio.writequery import CypherQuery

missing_types = {int:  np.inf, float: np.inf, str: '<MISSING>', type(None): np.inf, None: np.inf, bool: np.inf, datetime: datetime(1900, 1, 1, 0, 0)}
convert_types = {int:  float, float: float, str: str, type(None): float, None: float, bool: float,
                 datetime: lambda x: datetime(x.year, x.month, x.day, x.hour, x.minute, x.second),
                 list: list, tuple: tuple, np.ndarray: np.ndarray, pd.DataFrame: pd.DataFrame, pd.Series: pd.Series,
                 dict: dict, Table: Table}

def is_null(x):
    try:
        return np.all(x != x)
    except TypeError:
        pass
    return x is None


def _convert_datatypes(x, nan2missing=True, none2missing=True, surrounding_type=None):
    if isinstance(x, (tuple, list)):
        types = set(convert_types[type(i)] for i in x if not is_null(i))
        if len(types) == 0:
            acceptable_type = str
        elif len(types) == 1:
            acceptable_type = types.pop()
        else:
            raise TypeError(f"Lists must be of homogeneous type")
        r = [_convert_datatypes(xi, nan2missing, none2missing, surrounding_type=acceptable_type) for xi in x]
        if isinstance(x, tuple):
            r = tuple(r)
        return r
    elif isinstance(x, dict):
        return {_convert_datatypes(k, nan2missing, none2missing): _convert_datatypes(v, nan2missing, none2missing) for k, v in x.items()}
    elif isinstance(x, (pd.DataFrame, pd.Series)):
        return _convert_datatypes(pd.DataFrame(x).reset_index().to_dict('records'), nan2missing, none2missing)
    elif isinstance(x, Table):
        x = list(map(lambda row: {c: ri for c, ri in zip(x.colnames, row)}, x.iterrows()))
        return _convert_datatypes(pd.DataFrame(x).reset_index().to_dict('records'), nan2missing, none2missing)
    elif isinstance(x, np.ndarray):
        return _convert_datatypes(x.tolist(), nan2missing, none2missing)
    if not (none2missing or nan2missing):
        return x
    elif none2missing:
        if x is None:
            return missing_types[surrounding_type]
    if nan2missing:
        try:
            if np.isnan(x):
                return missing_types[surrounding_type]
        except TypeError:
            pass
    if surrounding_type is not None:
        return convert_types[surrounding_type](x)
    return x


class Graph(metaclass=ContextMeta):
    def __new__(cls, *args, **kwargs):
        # resolves the parent instance
        instance = super().__new__(cls)
        if kwargs.get("graph") is not None:
            instance._parent = kwargs.get("graph")
        else:
            instance._parent = cls.get_context(error_if_none=False)
        return instance

    def __init__(self, profile=None, name=None, **settings):
        self.write_allowed = settings.pop('write', False)
        self.neograph = NeoGraph(profile, name, **settings)

    def create_unique_constraint(self, label, key):
        try:
            self.neograph.schema.create_uniqueness_constraint(label, key)
        except py2neo.database.work.ClientError:
            pass

    def drop_unique_constraint(self, label, key):
        try:
            self.neograph.schema.drop_uniqueness_constraint(label, key)
        except py2neo.database.work.DatabaseError:
            pass

    def write(self, collision_manager):
        return CypherQuery(collision_manager)

    def _execute(self, cypher, parameters, backoff=1, limit=10):
        try:
            return self.neograph.run(cypher, parameters=parameters)
        except RuntimeError as e:
            if backoff >= limit:
                raise e
            warn(f'Connection possibly busy, waiting {backoff} seconds to retry. Actual error was {e}')
            sleep(backoff)
            backoff *= 2
            return self._execute(cypher, parameters, backoff, limit)

    def execute(self, cypher, **payload):
        if not self.write_allowed:
            raise IOError(f"Write is not allowed, set `write=True` to permit writing.")
        d = _convert_datatypes(payload, nan2missing=True, none2missing=True)
        return self._execute(cypher, d)

    def output_for_debug(self, **payload):
        d = _convert_datatypes(payload, nan2missing=True, none2missing=True)
        warn(f"When parameters are output for debug in the neo4j desktop, it cannot be guaranteed the data types will remain the same. "
             f"For certain, infs/nans/None are converted to strings (to avoid this, run your query without using `output_for_debug`)")
        return f':params {d}'

Graph._context_class = Graph



from pathlib import Path

import os

from weaveio.config_tables import progtemp_config
from weaveio.hierarchy import Hierarchy, Multiple, Indexed, One2One

HERE = Path(os.path.dirname(os.path.abspath(__file__)))


class Author(Hierarchy):
    is_template = True


class CASU(Author):
    idname = 'casuid'


class APS(Author):
    idname = 'apsvers'


class Simulator(Author):
    factors = ['simvdate', 'simver', 'simmode']
    identifier_builder = factors


class System(Author):
    idname = 'sysver'


class ArmConfig(Hierarchy):
    factors = ['resolution', 'vph', 'camera', 'colour']
    identifier_builder = ['resolution', 'vph', 'camera']

    def __init__(self, tables=None, **kwargs):
        if kwargs['vph'] == 3 and kwargs['camera'] == 'blue':
            kwargs['colour'] = 'green'
        else:
            kwargs['colour'] = kwargs['camera']
        super().__init__(tables, **kwargs)

    @classmethod
    def from_progtemp_code(cls, progtemp_code):
        config = progtemp_config.loc[progtemp_code[0]]
        red = cls(resolution=str(config.resolution), vph=int(config.red_vph), camera='red')
        blue = cls(resolution=str(config.resolution), vph=int(config.blue_vph), camera='blue')
        return red, blue


class ObsTemp(Hierarchy):
    factors = ['maxseeing', 'mintrans', 'minelev', 'minmoon', 'maxsky', 'code']
    identifier_builder = factors[:-1]

    @classmethod
    def from_header(cls, header):
        names = [f.lower() for f in cls.factors[:-1]]
        obstemp_code = list(header['OBSTEMP'])
        return cls(**{n: v for v, n in zip(obstemp_code, names)}, code=header['OBSTEMP'])


class Survey(Hierarchy):
    idname = 'name'


class WeaveTarget(Hierarchy):
    idname = 'cname'


class Fibre(Hierarchy):
    idname = 'fibreid'


class SubProgramme(Hierarchy):
    parents = [Multiple(Survey)]
    factors = ['name']
    idname = 'progid'


class SurveyCatalogue(Hierarchy):
    parents = [SubProgramme]
    factors = ['name']
    idname = 'catid'


class SurveyTarget(Hierarchy):
    parents = [SurveyCatalogue, WeaveTarget]
    factors = ['targid', 'targname', 'targra', 'targdec', 'targepoch',
               'targpmra', 'targpmdec', 'targparal', 'mag_g', 'emag_g', 'mag_r', 'emag_r', 'mag_i', 'emag_i', 'mag_gg', 'emag_gg',
               'mag_bp', 'emag_bp', 'mag_rp', 'emag_rp']
    identifier_builder = ['weavetarget', 'surveycatalogue', 'targid', 'targra', 'targdec']


class InstrumentConfiguration(Hierarchy):
    factors = ['mode', 'binning']
    parents = [Multiple(ArmConfig, 2, 2, idname='camera')]
    identifier_builder = ['armconfigs', 'mode', 'binning']


class ProgTemp(Hierarchy):
    parents = [InstrumentConfiguration]
    factors = ['length', 'exposure_code', 'code']
    identifier_builder = ['instrumentconfiguration'] + factors

    @classmethod
    def from_progtemp_code(cls, progtemp_code):
        progtemp_code = progtemp_code.split('.')[0]
        progtemp_code_list = list(map(int, progtemp_code))
        configs = ArmConfig.from_progtemp_code(progtemp_code_list)
        mode = progtemp_config.loc[progtemp_code_list[0]]['mode']
        binning = progtemp_code_list[3]
        config = InstrumentConfiguration(armconfigs=configs, mode=mode, binning=binning)
        exposure_code = progtemp_code[2:4]
        length = progtemp_code_list[1]
        return cls(code=progtemp_code, length=length, exposure_code=exposure_code,
                   instrumentconfiguration=config)


class OBSpec(Hierarchy):
    factors = ['obtitle']
    parents = [ObsTemp, ProgTemp, Multiple(SurveyCatalogue), Multiple(SubProgramme), Multiple(Survey)]
    idname = 'xml'  # this is CAT-NAME in the header not CATNAME, annoyingly no hyphens allowed


class FibreTarget(Hierarchy):
    factors = ['fibrera', 'fibredec', 'status', 'xposition', 'yposition',
               'orientat',  'retries', 'targx', 'targy', 'targuse', 'targprio']
    parents = [OBSpec, Fibre, SurveyTarget]
    identifier_builder = ['obspec', 'fibre', 'surveytarget', 'fibrera', 'fibredec', 'targuse']
    belongs_to = ['obspec', 'surveytarget']


class OB(Hierarchy):
    idname = 'obid'  # This is globally unique by obid
    factors = ['obstartmjd']
    parents = [OBSpec]


class Exposure(Hierarchy):
    idname = 'expmjd'  # globally unique
    parents = [OB]


class Run(Hierarchy):
    idname = 'runid'
    parents = [ArmConfig, Exposure]


class Observation(Hierarchy):
    parents = [One2One(Run), CASU, Simulator, System]
    factors = ['mjdobs', 'seeing', 'windspb', 'windspe', 'humidb', 'humide', 'winddir', 'airpres', 'tempb', 'tempe', 'skybrght', 'observer']
    products = {'primary': 'primary', 'guidinfo': 'guidinfo', 'metinfo': 'metinfo'}
    identifier_builder = ['run', 'mjdobs']
    version_on = ['run']

    @classmethod
    def from_header(cls, run, header):
        factors = {f: header.get(f) for f in cls.factors}
        factors['mjdobs'] = float(header['MJD-OBS'])
        casu = CASU(casuid=header.get('casuvers', header.get('casuid')))
        sim = Simulator(simver=header['simver'], simmode=header['simmode'], simvdate=header['simvdate'])
        sys = System(sysver=header['sysver'])
        return cls(run=run, casu=casu, simulator=sim, system=sys, **factors)


class SourcedData(Hierarchy):
    is_template = True
    factors = ['sourcefile', 'nrow']
    identifier_builder = ['sourcefile', 'nrow']


class Spectrum(SourcedData):
    is_template = True
    plural_name = 'spectra'


class RawSpectrum(Spectrum):
    plural_name = 'rawspectra'
    parents = [One2One(Observation), CASU]
    products = {'counts1': 'counts1', 'counts2': 'counts2'}
    version_on = ['observation']
    # any duplicates under a run will be versioned based on their appearance in the database
    # only one raw per run essentially


class L1SpectrumRow(Spectrum):
    plural_name = 'l1spectrumrows'
    is_template = True
    products = {'primary': 'primary', 'flux': Indexed('flux'), 'ivar': Indexed('ivar'),
                'flux_noss': Indexed('flux_noss'), 'ivar_noss': Indexed('ivar_noss'), 'sensfunc': 'sensfunc'}
    factors = Spectrum.factors + ['nspec', 'exptime', 'snr', 'meanflux_g', 'meanflux_r', 'meanflux_i', 'meanflux_gg', 'meanflux_bp', 'meanflux_rp']


class L1SingleSpectrum(L1SpectrumRow):
    plural_name = 'l1singlespectra'
    parents = L1SpectrumRow.parents + [RawSpectrum, FibreTarget, CASU]
    version_on = ['rawspectrum', 'fibretarget']
    factors = L1SpectrumRow.factors + [
        'rms_arc1', 'rms_arc2', 'resol', 'helio_cor',
        'wave_cor1', 'wave_corrms1', 'wave_cor2', 'wave_corrms2',
        'skyline_off1', 'skyline_rms1', 'skyline_off2', 'skyline_rms2',
        'sky_shift', 'sky_scale']


class L1StackSpectrum(L1SpectrumRow):
    plural_name = 'l1stackspectra'
    parents = L1SpectrumRow.parents + [Multiple(L1SingleSpectrum, 2), OB, ArmConfig, FibreTarget, CASU]
    version_on = ['l1singlespectra', 'fibretarget']


class L1SuperStackSpectrum(L1SpectrumRow):
    plural_name = 'l1superstackspectra'
    parents = L1SpectrumRow.parents + [Multiple(L1SingleSpectrum, 2), OBSpec, ArmConfig, FibreTarget, CASU]
    version_on = ['l1singlespectra', 'fibretarget']


class L1SuperTargetSpectrum(L1SpectrumRow):
    plural_name = 'l1supertargetspectra'
    parents = L1SpectrumRow.parents + [Multiple(L1SingleSpectrum, 2), WeaveTarget, CASU]
    version_on = ['l1singlespectra', 'weavetarget']


class L2(SourcedData):
    is_template = True


class L2Single(L2):
    parents = [Multiple(L1SingleSpectrum, 2, 3), FibreTarget, APS, Exposure]


class L2Stack(L2):
    parents = [Multiple(L1SingleSpectrum, 0, 3), Multiple(L1StackSpectrum, 0, 3), FibreTarget, APS, OB]


class L2SuperStack(L2):
    parents = [Multiple(L1SingleSpectrum, 0, 3), Multiple(L1StackSpectrum, 0, 3), Multiple(L1SuperStackSpectrum, 0, 3), FibreTarget, APS, OBSpec]


class L2SuperTarget(L2):
    parents = [Multiple(L1SuperTargetSpectrum, 2, 3), APS, WeaveTarget]


class L2SourcedData(Hierarchy):
    is_template = True
    factors = ['sourcefile', 'hduname', 'nrow']
    identifier_builder = ['sourcefile', 'hduname', 'nrow']
    parents = [One2One(L2)]
    belongs_to = ['l2']


class L2TableRow(L2SourcedData):
    is_template = True


class L2Spectrum(L2SourcedData):
    is_template = True
    plural_name = 'l2spectra'
    products = {'flux': Indexed('*_spectra', 'flux'), 'ivar': Indexed('*_spectra', 'ivar'),
                'model_ab': Indexed('*_spectra', 'model_ab'), 'model_em': Indexed('*_spectra', 'model_em'),
                'lambda': Indexed('*_spectra', 'lambda')}


class ClassificationTable(L2TableRow):
    factors = L2TableRow.factors + ['class', 'subclass', 'z', 'z_err', 'auto_class_alls',
                                    'auto_subclass_alls', 'z_alls', 'z_err_alls', 'rchi2diff',
                                    'rchi2_alls', 'rchi2diff_alls', 'zwarning', 'zwarning_alls',
                                    'sn_median_all', 'sn_medians', 'specflux_sloans',
                                    'specflux_sloan_ivars', 'specflux_johnsons',
                                    'specflux_johnson_ivars', 'specsynfluxes', 'specsynflux_ivars',
                                    'specskyflux']


class GalaxyTable(L2TableRow):
    with open(HERE / 'galaxy_table_columns.txt', 'r') as _f:
        factors = L2TableRow.factors + [x.lower().strip() for x in _f.readlines() if len(x)]


class ClassificationSpectrum(L2Spectrum):
    plural_name = 'classification_spectra'
    products = {'flux': Indexed('class_spectra', 'flux'), 'ivar': Indexed('class_spectra', 'ivar'),
                'model': Indexed('class_spectra', 'model'), 'lambda': Indexed('class_spectra', 'lambda')}


class GalaxySpectrum(L2Spectrum):
    plural_name = 'galaxy_spectra'
    products = {'flux': Indexed('galaxy_spectra', 'flux'), 'ivar': Indexed('galaxy_spectra', 'ivar'),
                'model_ab': Indexed('galaxy_spectra', 'model_ab'), 'model_em': Indexed('galaxy_spectra', 'model_em'),
                'lambda': Indexed('galaxy_spectra', 'lambda')}



from collections import defaultdict
from functools import reduce
from operator import xor
from textwrap import dedent
from typing import List, Callable, Union

import re
from warnings import warn

from ..context import ContextMeta


def camelcase(x):
    if x[0] != x[0].upper():
        raise ValueError(f"Labels must be camelcase")
    return x


class CypherQuery(metaclass=ContextMeta):
    def __init__(self, collision_manager='track&flag'):
        self.data = []
        self.statements = [TimeStamp()]  # type: List[BaseStatement]
        self.timestamp = self.statements[0].output_variables[0]
        self.open_contexts = [[self.timestamp]]
        self.closed_context = None
        self.collision_manager = collision_manager

    @property
    def current_context(self):
        return self.open_contexts[-1]

    @property
    def accessible_variables(self):
        return [v for context in self.open_contexts[::-1] for v in context] + [i for i in self.data]

    def is_accessible(self, variable):
        if variable in self.accessible_variables:
            return True
        if hasattr(variable, 'parent'):
            return self.is_accessible(variable.parent)
        return False

    def add_statement(self, statement, safe=True):
        for v in statement.input_variables:
            if safe:
                if not self.is_accessible(v):
                    raise ValueError(f"{v} is not accessible is this context. Have you left a WITH context?")
        self.statements.append(statement)
        self.current_context.extend(statement.output_variables)

    def remove_variable_names(self):
        for statement in self.statements:
            for v in statement.input_variables + statement.output_variables + statement.hidden_variables:
                v._name = None

    def make_variable_names(self):
        d = defaultdict(int)
        for data in self.data:
            namehint = data.namehint.lower()
            i = d[namehint]
            d[namehint] += 1
            data._name = f'{namehint}{i}'
        for statement in self.statements:
            for v in statement.input_variables + statement.output_variables + statement.hidden_variables:
                if v.name is None:
                    namehint = v.namehint.lower()
                    i = d[namehint]
                    d[namehint] += 1
                    v._name = f'{namehint}{i}'

    def render_query(self, procedure_tag=''):
        if not isinstance(self.statements[-1], Returns):
            self.returns(self.timestamp)
        self.make_variable_names()
        q = '\n'.join([s.to_cypher() for s in self.statements])
        datadict = {d.name: d.data for d in self.data}
        return dedent(re.sub(r'(custom\.[\w\d]+)\(', fr'\1-----{procedure_tag}(', q).replace('-----', '')), datadict

    def open_context(self):
        self.open_contexts.append([])

    def close_context(self):
        self.closed_context = self.open_contexts[-1]
        del self.open_contexts[-1]

    def add_data(self, data):
        self.data.append(data)

    def returns(self, *args, **kwargs):
        if not isinstance(self.statements[-1], Returns):
            self.statements.append(Returns(*args, **kwargs))
        else:
            self.statements[-1] = Returns(*args, **kwargs)
        return self


CypherQuery._context_class = CypherQuery


class Varname:
    def __init__(self, key):
        self.key = key

    def __repr__(self):
        return self.key

    def __eq__(self, other):
        return getattr(other, 'key', other) == self.key

    def __hash__(self):
        return hash(hash(self.key) + hash('varname'))


class BaseStatement:
    def __init__(self, input_variables, output_variables, hidden_variables=None):
        self.input_variables = list(input_variables)
        self.output_variables = list(output_variables)
        if hidden_variables is None:
            hidden_variables = []
        self.hidden_variables = hidden_variables

    def to_cypher(self):
        raise NotImplementedError

    def __eq__(self, other):
        return set(self.input_variables) == set(other.input_variables) \
               and set(self.output_variables) == set(other.output_variables) \
               and set(self.hidden_variables) == set(other.hidden_variables) \
               and self.__class__ is other.__class__

    def __hash__(self):
        return reduce(xor, map(hash, [tuple(self.input_variables), tuple(self.output_variables)]))


class Statement(BaseStatement):
    """A cypher statement that takes inputs and returns outputs"""
    def __init__(self, input_variables, output_variables, hidden_variables=None):
        super(Statement, self).__init__(input_variables, output_variables, hidden_variables)
        self.timestamp = CypherQuery.get_context().timestamp


class CustomStatement(Statement):
    def __init__(self, statement: Callable, input_variables, output_variables):
        super().__init__(input_variables, output_variables)
        self.statement = statement

    def to_cypher(self):
        return self.statement(*self.input_variables, *self.output_variables)


class Returns(Statement):
    def __init__(self, *unnamed_variables, **named_variables):
        self.unnamed_variables = unnamed_variables
        self.named_variables = named_variables
        self.input_variables = list(unnamed_variables) + list(named_variables.values())
        self.output_variables = []
        self.hidden_variables = []

    def to_cypher(self):
        l = list(map(str, self.unnamed_variables))
        l += [f'{v} as {k}' for k, v in self.named_variables.items()]
        return 'RETURN ' + ', '.join(l)


class TimeStamp(Statement):
    def __init__(self):
        self.input_variables = []
        self.output_variables = [CypherVariable('time')]
        self.hidden_variables = []

    def to_cypher(self):
        return f'WITH *, timestamp() as {self.output_variables[0]}'


class Alias(Statement):
    def __init__(self, input_thing, alias_namehint):
        self.out = CypherVariable(alias_namehint)
        self.input_thing = input_thing
        super(Alias, self).__init__([input_thing], [self.out])

    def to_cypher(self):
        return f"WITH *, {self.input_thing} as {self.out}"


class CypherVariable:
    def __init__(self, namehint=None):
        self.namehint = 'variable' if namehint is None else namehint
        self._name = None

    @property
    def name(self):
        return self._name

    def __repr__(self):
        if self.name is None:
            return super(CypherVariable, self).__repr__()
        return self.name

    def __getitem__(self, item: Union[str, int]):
        return self.get(item, alias=True)

    def get(self, item: Union[str, int], alias=False):
        assert isinstance(item, (int, str))
        getitem = CypherVariableItem(self, item)
        if alias:
            query = CypherQuery.get_context()
            if isinstance(item, int):
                item = f'{self.namehint}_index{item}'
            alias_statement = Alias(getitem, str(item))
            query.add_statement(alias_statement)
            return alias_statement.out
        else:
            return getitem


class DerivedCypherVariable(CypherVariable):
    def __init__(self, parent, args):
        super(DerivedCypherVariable, self).__init__()
        self.parent = parent
        self.args = args

    def string_formatter(self, parent, args):
        raise NotImplementedError

    @property
    def name(self):
        return self.string_formatter(self.parent, self.args)


class CypherVariableItem(DerivedCypherVariable):
    def string_formatter(self, parent, attr):
        if isinstance(attr, str):
            return f"{parent}['{attr}']"
        return f"{parent}[{attr}]"


class Collection(CypherVariable):
    pass


class CypherData(CypherVariable):
    def __init__(self, data, name='data', delay=False):
        super(CypherData, self).__init__(name)
        self.data = data
        if not delay:
            query = CypherQuery.get_context()  # type: CypherQuery
            query.add_data(self)

    def __repr__(self):
        return '$' + super(CypherData, self).__repr__()


import pytest
import py2neo


def test_merge_many_to_one_in_isolation(database: py2neo.Graph, procedure_tag: str):
    q = f"""
        UNWIND range(0, 100) as i
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes
        
        MERGE (b1: B {{id: 1}})
        
        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)
        
        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames
        
        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification
        
        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c // just to reset here
        MATCH (n:MyLabel)
        RETURN n
        """
    t = database.run(q).to_table()
    assert len(t) == 1


def test_does_not_duplicate_on_same(database: py2neo.Graph, procedure_tag: str):
    q = f"""
        UNWIND range(0, 100) as i
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c // just to reset here
        
        UNWIND range(0, 100) as i
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child

        MATCH (n:MyLabel)
        RETURN n
        """
    t = database.run(q).to_table()
    assert len(t) == 1



def test_creates_two_on_subset(database: py2neo.Graph, procedure_tag: str):
    q = f"""
        UNWIND range(0, 100) as i
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c // just to reset here
        
        UNWIND range(0, 10) as i  // make a subset
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c  // just to reset here

        UNWIND range(0, 100) as i  // do it again
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c // just to reset here

        MATCH (n:MyLabel)
        RETURN n
        """
    t = database.run(q).to_table()
    assert len(t) == 2



def test_creates_two_on_superset(database: py2neo.Graph, procedure_tag: str):
    q = f"""
        UNWIND range(0, 100) as i
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c // just to reset here
        
        UNWIND range(0, 200) as i  // make a superset
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c  // just to reset here

        UNWIND range(0, 100) as i  // do it again
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c // just to reset here

        MATCH (n:MyLabel)
        RETURN n
        """
    t = database.run(q).to_table()
    assert len(t) == 2



def test_creates_two_on_rel_change(database: py2neo.Graph, procedure_tag: str):
    q = f"""
        UNWIND range(0, 100) as i
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c // just to reset here
        
        UNWIND range(0, 100) as i  // make a superset
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel_different', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c  // just to reset here

        UNWIND range(0, 100) as i  // do it again
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c // just to reset here

        MATCH (n:MyLabel)
        RETURN n
        """
    t = database.run(q).to_table()
    assert len(t) == 2




def test_creates_two_on_prop_change(database: py2neo.Graph, procedure_tag: str):
    q = f"""
        UNWIND range(0, 100) as i
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c // just to reset here
        
        UNWIND range(0, 100) as i  // make a superset
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai, something_else: 1}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c  // just to reset here

        UNWIND range(0, 100) as i  // do it again
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c // just to reset here

        MATCH (n:MyLabel)
        RETURN n
        """
    t = database.run(q).to_table()
    assert len(t) == 2


def test_creates_one_to_one_correctly_on_spec_change(database: py2neo.Graph, procedure_tag: str):
    q = f"""
    MERGE (a: A {{id: 1}})
    WITH [[a, 'rel', {{order: 0}}]] as specs
    CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{match: 'match'}}, {{oncreate:1}}, {{oncreate: 1}}) YIELD child
    
    MERGE (a: A {{id: 1}})
    WITH [[a, 'rel', {{order: 0}}]] as specs
    CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{match: 'match'}}, {{oncreate: 2}}, {{oncreate: 2}}) YIELD child

    MERGE (a: A {{id: 2}})
    WITH [[a, 'rel', {{order: 0}}]] as specs
    CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{match: 'match'}}, {{oncreate:3}}, {{oncreate: 3}}) YIELD child

    MATCH (n:MyLabel)
    MATCH (n)<-[r]-(a:A)
    RETURN n.oncreate, r.oncreate
    ORDER BY n.oncreate
    """
    t = database.run(q).to_table()
    assert len(t) == 2
    assert t[0][0] == 1
    assert t[0][1] == 1
    assert t[1][0] == 3
    assert t[1][1] == 3


def test_creates_one_to_one_correctly_on_child_change(database: py2neo.Graph, procedure_tag: str):
    q = f"""
    MERGE (a: A {{id: 1}})
    WITH [[a, 'rel', {{order: 0}}]] as specs
    CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{match: 'match'}}, {{oncreate:1}}, {{oncreate: 1}}) YIELD child
    
    MERGE (a: A {{id: 1}})
    WITH [[a, 'rel', {{order: 0}}]] as specs
    CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{match: 'match'}}, {{oncreate: 2}}, {{oncreate: 2}}) YIELD child

    MERGE (a: A {{id: 1}})
    WITH [[a, 'rel', {{order: 0}}]] as specs
    CALL custom.multimerge{procedure_tag}(specs, ['MyLabel', 'MyOtherLabel'], {{match: 'match'}}, {{oncreate:3}}, {{oncreate: 3}}) YIELD child

    MATCH (n:MyLabel)
    MATCH (n)<-[r]-(a:A)
    RETURN n.oncreate, r.oncreate
    ORDER BY n.oncreate
    """
    t = database.run(q).to_table()
    assert len(t) == 2
    assert t[0][0] == 1
    assert t[0][1] == 1
    assert t[1][0] == 3
    assert t[1][1] == 3


def test_creates_three_on_super_and_subsets(database: py2neo.Graph, procedure_tag: str):
    q = f"""
        UNWIND range(0, 100) as i
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c // just to reset here
        
        UNWIND range(0, 200) as i  // make a superset
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c  // just to reset here

        UNWIND range(0, 10) as i  // make subset
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c // just to reset here

        MATCH (n:MyLabel)
        RETURN n
        """
    t = database.run(q).to_table()
    assert len(t) == 3


@pytest.mark.parametrize('beforelabels,afterlabels', [
    ('["Label1"]', '["Label1"]'),
    ('["Label1"]', '["Label2"]'),
    ('["Label1"]', '["Label1", "Label2"]'),
    ('["Label1", "Label2"]', '["Label1"]'),
])
def test_creates_new_when_label_changes(database: py2neo.Graph, procedure_tag: str,
                                        beforelabels, afterlabels):
    q = f"""
        UNWIND range(0, 100) as i
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, {beforelabels}, {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c // just to reset here
        
        UNWIND range(0, 100) as i  // make a superset
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, {afterlabels}, {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c  // just to reset here

        MATCH (n)
        WHERE NOT n:A AND NOT n:B
        RETURN n
        """
    t = database.run(q).to_table()
    if beforelabels == afterlabels:
        assert len(t) == 1
    else:
        assert len(t) == 2


@pytest.mark.parametrize('beforeproperties,afterproperties,diff', [
    ('{a: 1}', '{a: 1}', False),
    ('{a: 1}', '{a: 2}', True),
    ('{a: 1}', '{a: 1, b: 2}', True),  # we merge based on given properties only
    ('{}', '{}', False),
    ('{}', '{a: 1}', True),
    ('{a:1}', '{}', False),  # we merge based on given properties only
    ('{a:1, b: 2}', '{a: 1}', False),  # we merge based on given properties only
])
def test_creates_new_when_properties_change(database: py2neo.Graph, procedure_tag: str,
                                            beforeproperties, afterproperties, diff):
    q = f"""
        UNWIND range(0, 100) as i
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {beforeproperties}, {{}}, {{}}) YIELD child
        WITH collect(child) as c // just to reset here
        
        UNWIND range(0, 100) as i  // make a superset
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {afterproperties}, {{}}, {{}}) YIELD child
        WITH collect(child) as c  // just to reset here

        MATCH (n:MyLabel)
        RETURN n
        """
    t = database.run(q).to_table()
    if not diff:
        assert len(t) == 1
    else:
        assert len(t) == 2



def test_createprops_rels_are_created(database: py2neo.Graph, procedure_tag: str):
    q = f"""
        UNWIND range(0, 100) as i
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{match: 1}}, {{create: 'node1'}}, {{create: 'rel1'}}) YIELD child
        WITH collect(child) as c // just to reset here

        UNWIND range(0, 200) as i  // make a superset
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{match: 2}}, {{create: 'node2'}}, {{create: 'rel2'}}) YIELD child
        WITH collect(child) as c  // just to reset here

        UNWIND range(0, 100) as i  // and again
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{match: 1}}, {{create: 'node3'}}, {{create: 'rel3'}}) YIELD child
        WITH collect(child) as c // just to reset here

        MATCH (n:MyLabel)
        MATCH (other)-[r]->(n)
        RETURN n, collect(n.create) as created, collect(n.match) as matched, collect(r.create) as createdrel
        ORDER BY n.match
        """
    t = database.run(q).to_table()
    assert len(t) == 2
    assert all(i == 'node1' for i in t[0][1])  # create node prop
    assert all(i == 1 for i in t[0][2])  # match
    assert all(i == 'rel1' for i in t[0][3])  # create rel prop
    #
    assert all(i == 'node2' for i in t[1][1])  # create node prop
    assert all(i == 2 for i in t[1][2])  # match
    assert all(i == 'rel2' for i in t[1][3])  # create rel prop


def test_leaving_properties_out_returns_all_children_without_modification(database: py2neo.Graph, procedure_tag: str):
    q = f"""
        UNWIND range(0, 100) as i
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c // just to reset here

        UNWIND range(0, 10) as i  // make a subset
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{match: 1}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c  // just to reset here

        UNWIND range(0, 100) as i  // do it again
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{match: 2}}, {{}}, {{}}) YIELD child
        WITH collect(child) as c, specs // just to reset here

        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        RETURN child.match
        ORDER BY child.match
        """
    t = database.run(q).to_table()
    assert len(t) == 2  # on the ones with 100 A nodes
    assert t[0][0] == 2
    assert t[1][0] is None  # these are

    assert database.run('MATCH (n: MyLabel) RETURN count(n)').to_table()[0][0] == 3


import inspect
from functools import wraps, partial
from typing import Tuple, Dict, Type, Union, List, Optional
from warnings import warn

from . import writequery
from .writequery import CypherQuery, Unwind, Collection, CypherVariable
from .context import ContextError
from .utilities import Varname

def _convert_types_to_node(x):
    if isinstance(x, dict):
        return {_convert_types_to_node(k): _convert_types_to_node(v) for k, v in x.items()}
    elif isinstance(x, (list, set, tuple)):
        return x.__class__([_convert_types_to_node(i) for i in x])
    elif isinstance(x, Graphable):
        return x.node
    else:
        return x

def hierarchy_query_decorator(function):
    @wraps(function)
    def inner(*args, **kwargs):
        args = _convert_types_to_node(args)
        kwargs = _convert_types_to_node(kwargs)
        return function(*args, **kwargs)
    return inner


unwind = hierarchy_query_decorator(writequery.unwind)
merge_node = hierarchy_query_decorator(writequery.merge_node)
match_node = hierarchy_query_decorator(writequery.match_node)
match_pattern_node = hierarchy_query_decorator(writequery.match_pattern_node)
collect = hierarchy_query_decorator(writequery.collect)
merge_relationship = hierarchy_query_decorator(writequery.merge_relationship)
set_version = hierarchy_query_decorator(writequery.set_version)


def chunker(lst, n):
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


FORBIDDEN_LABELS = []
FORBIDDEN_PROPERTY_NAMES = []
FORBIDDEN_LABEL_PREFIXES = ['_']
FORBIDDEN_PROPERTY_PREFIXES = ['_']
FORBIDDEN_IDNAMES = ['idname']


class RuleBreakingException(Exception):
    pass


class Multiple:
    def __init__(self, node, minnumber=1, maxnumber=None, constrain=None, idname=None):
        self.node = node
        self.minnumber = minnumber
        self.maxnumber = maxnumber
        self.name = node.plural_name
        self.singular_name = node.singular_name
        self.plural_name = node.plural_name
        self.idname = self.node.idname
        self.relation_idname = idname
        try:
            self.factors =  self.node.factors
        except AttributeError:
            self.factors = []
        try:
            self.parents = self.node.parents
        except AttributeError:
            self.parents = []
        self.constrain = [] if constrain is None else constrain

    def __repr__(self):
        return f"<Multiple({self.node} [{self.minnumber} - {self.maxnumber}])>"


class One2One(Multiple):
    def __init__(self, node, constrain=None, idname=None):
        super(One2One, self).__init__(node, 1, 1, constrain, idname)

    def __repr__(self):
        return f"<One2One({self.node})>"


class Indexed:
    def __init__(self, hdu_name, column_name=None):
        self.name = hdu_name
        self.column_name = column_name


class GraphableMeta(type):
    def __new__(meta, name: str, bases, _dct):
        dct = {'is_template': False}
        dct.update(_dct)
        dct['aliases'] = dct.get('aliases', [])
        dct['aliases'] += [a for base in bases for a in base.aliases]
        if dct.get('plural_name', None) is None:
            dct['plural_name'] = name.lower() + 's'
        dct['singular_name'] = name.lower()
        if dct['plural_name'] != dct['plural_name'].lower():
            raise RuleBreakingException(f"plural_name must be lowercase")
        if dct['singular_name'] != dct['singular_name'].lower():
            raise RuleBreakingException(f"singular_name must be lowercase")
        idname = dct.get('idname', None)
        if idname in FORBIDDEN_IDNAMES:
            raise RuleBreakingException(f"You may not name an id as one of {FORBIDDEN_IDNAMES}")
        if not (isinstance(idname, str) or idname is None):
            raise RuleBreakingException(f"{name}.idname ({idname}) must be a string or None")
        if name[0] != name.capitalize()[0] or '_' in name:
            raise RuleBreakingException(f"{name} must have `CamelCaseName` style name")
        for factor in dct.get('factors', []) + ['idname'] + [dct['singular_name'], dct['plural_name']]:
            if factor != factor.lower():
                raise RuleBreakingException(f"{name}.{factor} must have `lower_snake_case` style name")
            if factor in FORBIDDEN_PROPERTY_NAMES:
                raise RuleBreakingException(f"The name {factor} is not allowed for class {name}")
            if any(factor.startswith(p) for p in FORBIDDEN_PROPERTY_PREFIXES):
                raise RuleBreakingException(f"The name {factor} may not start with any of {FORBIDDEN_PROPERTY_PREFIXES} for {name}")
        r = super(GraphableMeta, meta).__new__(meta, name, bases, dct)
        return r

    def __init__(cls, name, bases, dct):
        if cls.idname is not None and cls.identifier_builder is not None:
            raise RuleBreakingException(f"You cannot define a separate idname and an identifier_builder at the same time for {name}")
        if cls.indexes and (cls.idname is not None or cls.identifier_builder is not None):
            raise RuleBreakingException(f"You cannot define an index and an id at the same time for {name}")
        nparents_in_id = 0
        parentnames = {}
        for i in cls.parents:
            if isinstance(i, One2One):
                parentnames[i.singular_name] = (1, 1)
            elif isinstance(i, Multiple):
                parentnames[i.plural_name] = (i.minnumber, i.maxnumber)
            else:
                parentnames[i.singular_name] = (1, 1)
        if cls.identifier_builder is not None:
            for p in cls.identifier_builder:
                if p in parentnames:
                    mn, mx = parentnames[p]
                    if mn == 0:
                        raise RuleBreakingException(f"Cannot make an id from an optional (min=0) parent for {name}")
                    if mx != mn:
                        raise RuleBreakingException(f"Cannot make an id from an unbound (max!=min) parent for {name}")
                    nparents_in_id += mx
                elif p in cls.factors:
                    pass
                else:
                    raise RuleBreakingException(f"Unknown identifier source {p} for {name}")
        version_parents = []
        version_factors = []
        for i in cls.version_on:
            parents = [p.node if isinstance(p, One2One) else p for p in cls.parents]
            if i in [p.singular_name if isinstance(p, type) else p.name for p in parents]:
                version_parents.append(i)
            elif i in cls.factors:
                version_factors.append(i)
            else:
                raise RuleBreakingException(f"Unknown {i} to version on for {name}. Must refer to a parent or factor.")
        if len(version_factors) > 1 and len(version_parents) == 0:
            raise RuleBreakingException(f"Cannot build a version relative to nothing. You must version on at least one parent.")
        if not cls.is_template:
            if not (len(cls.indexes) or cls.idname or
                    (cls.identifier_builder is not None and len(cls.identifier_builder) > 0)):
                raise RuleBreakingException(f"{name} must define an indexes, idname, or identifier_builder")
        for i in cls.indexes:
            if i not in cls.parents and i not in cls.factors:
                raise RuleBreakingException(f"index {i} of {name} must be a factor or parent of {name}")
        if len(cls.hdus):
            hduclasses = {}
            for i, (hduname, hdu) in enumerate(cls.hdus.items()):
                if hdu is not None:
                    typename = name+hduname[0].upper()+hduname[1:]
                    typename = typename.replace('_', '')
                    hduclass = type(typename, (hdu, ), {'parents': [One2One(cls)], 'identifier_builder': [cls.singular_name, 'extn', 'name']})
                    hduclasses[hduname] = hduclass
                    if hduname in cls.factors or hduname in [p.singular_name if isinstance(p, type) else p.name for p in cls.parents]:
                        raise RuleBreakingException(f"There is already a factor/parent called {hduname} defined in {name}")
                    for base in bases:
                        if (hduname in base.factors or hduname in base.parents or hasattr(base, hduname)) and hduname not in base.hdus:
                            raise RuleBreakingException(f"There is already a factor/parent called {hduname} defined in {base}->{name}")
                    setattr(cls, hduname, hduclass)  # add as an attribute
            cls.hdus = hduclasses  # overwrite hdus
        if cls.concatenation_constants is not None:
            if len(cls.concatenation_constants):
                cls.factors = cls.factors + cls.concatenation_constants + ['concatenation_constants']
        clses = [i.__name__ for i in inspect.getmro(cls)]
        clses = clses[:clses.index('Graphable')]
        cls.neotypes = clses
        cls.products_and_factors = cls.factors + list(cls.products.keys())
        if cls.idname is not None:
            cls.products_and_factors.append(cls.idname)
        super().__init__(name, bases, dct)


class Graphable(metaclass=GraphableMeta):
    idname = None
    identifier = None
    indexer = None
    type_graph_attrs = {}
    plural_name = None
    singular_name = None
    parents = []
    uses_tables = False
    factors = []
    data = None
    query = None
    is_template = True
    products = {}
    indexes = []
    identifier_builder = None
    version_on = []
    hdus = {}
    produces = []
    concatenation_constants = []
    belongs_to = []
    products_and_factors = []

    @property
    def node(self):
        return self._node

    @node.setter
    def node(self, value):
        assert isinstance(value, CypherVariable)
        self._node = value

    @classmethod
    def requirement_names(cls):
        l = []
        for p in cls.parents:
            if isinstance(p, type):
                if issubclass(p, Graphable):
                    l.append(p.singular_name)
            else:
                if isinstance(p, One2One):
                    l.append(p.singular_name)
                if isinstance(p, Multiple):
                    l.append(p.plural_name)
                else:
                    raise RuleBreakingException(f"The parent list of a Hierarchy must contain "
                                                f"only other Hierarchies or Multiple(Hierarchy)")
        return l

    def add_parent_data(self, data):
        self.data = data

    def add_parent_query(self, query):
        self.query = query

    def __getattr__(self, item):
        if self.query is not None:
            warn('Lazily loading a hierarchy attribute can be costly. Consider using a more flexible query.')
            attribute = getattr(self.query, item)()
            setattr(self, item, attribute)
            return attribute
        raise AttributeError(f"Query not added to {self}, cannot search for {self}.{item}")

    @property
    def neoproperties(self):
        identifier_builder = [] if self.identifier_builder is None else self.identifier_builder
        d = {}
        for f in self.factors:
            if f not in identifier_builder and f != self.idname:
                value = getattr(self, f.lower())
                if value is not None:
                    d[f.lower()] = value
        return d

    @property
    def neoidentproperties(self):
        identifier_builder = [] if self.identifier_builder is None else self.identifier_builder
        d = {}
        if self.identifier is None and self.idname is not None:
            raise ValueError(f"{self} must have an identifier")
        if self.idname is None and self.identifier is not None:
            raise ValueError(f"{self} must have an idname to be given an identifier")
        elif self.idname is not None:
            d[self.idname] = self.identifier
            d['id'] = self.identifier
        for f in self.factors:
            if f in identifier_builder:
                value = getattr(self, f.lower())
                if value is not None:
                    d[f.lower()] = value
        return d


    def __init__(self, do_not_create=False, **predecessors):
        self.predecessors = predecessors
        self.data = None
        if do_not_create:
            return
        try:
            query = CypherQuery.get_context()  # type: CypherQuery
            collision_manager = query.collision_manager
        except ContextError:
            return
        merge_strategy = self.__class__.merge_strategy()
        version_parents = []
        if  merge_strategy == 'NODE FIRST':
            self.node = child = merge_node(self.neotypes, self.neoidentproperties, self.neoproperties,
                                           collision_manager=collision_manager)
            for k, parent_list in predecessors.items():
                type = 'is_required_by'
                if isinstance(parent_list, Collection):
                    with unwind(parent_list, enumerated=True) as (parent, i):
                        props = {'order': i}
                        merge_relationship(parent, child, type, {}, props, collision_manager=collision_manager)
                    parent_list = collect(parent)
                    if k in self.version_on:
                        raise RuleBreakingException(f"Cannot version on a collection of nodes")
                else:
                    for parent in parent_list:
                        props = {'order': 0}
                        merge_relationship(parent, child, type, {}, props, collision_manager=collision_manager)
                        if k in self.version_on:
                            version_parents.append(parent)
        elif merge_strategy == 'NODE+RELATIONSHIP':
            parentnames = [p.singular_name if isinstance(p, One2One) else p.plural_name if isinstance(p, Multiple) else p.singular_name for p in self.parents]
            parents = []
            others = []
            for k, parent_list in predecessors.items():
                if isinstance(parent_list, Collection):
                    raise TypeError(f"Cannot merge NODE+RELATIONSHIP for collections")
                if k in parentnames and k in self.identifier_builder:
                    parents += [p for p in parent_list]
                else:
                    others += [(i, p) for i, p in enumerate(parent_list)]
                if k in self.version_on:
                    version_parents += parent_list
            reltype = 'is_required_by'
            relparents = {p: (reltype, {'order': 0}, {}) for p in parents}
            child = self.node = merge_node(self.neotypes, self.neoidentproperties, self.neoproperties,
                                           parents=relparents, collision_manager=collision_manager)
            for i, other in others:
                merge_relationship(other, child, reltype, {'order': i}, {}, collision_manager=collision_manager)
        else:
            ValueError(f"Merge strategy not known: {merge_strategy}")
        if len(version_parents):
            version_factors = {f: self.neoproperties[f] for f in self.version_on if f in self.factors}
            set_version(version_parents, ['is_required_by'] * len(version_parents), self.neotypes[-1], child, version_factors)

    @classmethod
    def has_factor_identity(cls):
        if cls.identifier_builder is None:
            return False
        if len(cls.identifier_builder) == 0:
            return False
        for p in cls.parents:
            if isinstance(p, One2One):
                if p.singular_name in cls.identifier_builder:
                    return False
            elif isinstance(p, Multiple):
                if p.plural_name in cls.identifier_builder:
                    return False
            elif p.singular_name in cls.identifier_builder:
                return False
        return True

    @classmethod
    def has_parent_identity(cls):
        if cls.identifier_builder is None:
            return False
        if len(cls.identifier_builder) == 0:
            return False
        for p in cls.parents:
            if isinstance(p, One2One):
                if p.singular_name in cls.identifier_builder:
                    return True
            elif isinstance(p, Multiple):
                if p.plural_name in cls.identifier_builder:
                    return True
            elif p.singular_name in cls.identifier_builder:
                return True
        return False

    @classmethod
    def make_schema(cls) -> Optional[str]:
        name = cls.__name__
        if cls.idname is not None:
            prop = cls.idname
            return f'CREATE CONSTRAINT {name} ON (n:{name}) ASSERT (n.{prop}) IS NODE KEY'
        elif cls.identifier_builder:
            if cls.has_factor_identity():
                key = ', '.join([f'n.{f}' for f in cls.identifier_builder])
                return f'CREATE CONSTRAINT {name} ON (n:{name}) ASSERT ({key}) IS NODE KEY'
            elif cls.has_parent_identity():
                key = ', '.join([f'n.{f}' for f in cls.identifier_builder if f in cls.factors])
                if not len(key):
                    raise TypeError(f"No factors are present in the identity builder of {name} to make an index from ")
                return f'CREATE INDEX {name} FOR (n:{name}) ON ({key})'
        elif cls.indexes:
            key = ', '.join([f'n.{i}' for i in cls.indexes])
            return f'CREATE INDEX {name} FOR (n:{name}) ON ({key})'
        elif cls.is_template:
            return None
        else:
            raise RuleBreakingException(f"A hierarchy must define an idname, identifier_builder, or index, "
                                        f"unless it is marked as template class for something else (`is_template=True`)")

    @classmethod
    def merge_strategy(cls):
        if cls.idname is not None:
            return 'NODE FIRST'
        elif cls.identifier_builder:
            if cls.has_factor_identity():
                return 'NODE FIRST'
            elif cls.has_parent_identity():
                return 'NODE+RELATIONSHIP'
        return 'NODE FIRST'

    def attach_products(self, file=None, index=None, **hdus):
        """attaches products to a hierarchy with relations like: <-[:PRODUCT {index: rowindex, name: 'flux'}]-"""
        collision_manager = CypherQuery.get_context().collision_manager
        for productname, name in self.products.items():
            props = {}
            if isinstance(name, Indexed):
                if name.column_name is not None:
                    props['column_name'] = name.column_name
                name = name.name
                if index is None:
                    raise IndexError(f"{self} requires an index for {file} product {name}")
                props['index'] = index
            props['name'] = productname
            hdu = hdus[name]
            merge_relationship(hdu, self, 'product', props, {}, collision_manager=collision_manager)
        if file is not None:
            merge_relationship(file, self, 'is_required_by', {'name': 'file'}, {}, collision_manager=collision_manager)

    @classmethod
    def without_creation(cls, **kwargs):
        return cls(do_not_create=True, **kwargs)

    @classmethod
    def find(cls, anonymous_children=None, anonymous_parents=None, **kwargs):
        parent_names = [i.name if isinstance(i, Multiple) else i.singular_name for i in cls.parents]
        parents = [] if anonymous_parents is None else anonymous_parents
        anonymous_children = [] if anonymous_children is None else anonymous_children
        factors = {}
        for k, v in kwargs.items():
            if k in cls.factors:
                factors[k] = v
            elif k in parent_names:
                if not isinstance(v, list):
                    v = [v]
                for vi in v:
                    parents.append(vi)
            elif k == cls.idname:
                factors[k] = v
            else:
                raise ValueError(f"Unknown name {k} for {cls}")
        node = match_pattern_node(labels=cls.neotypes, properties=factors,
                                  parents=parents, children=anonymous_children)
        obj = cls.without_creation(**kwargs)
        obj.node = node
        return obj

    def __repr__(self):
        i = ''
        if self.idname is not None:
            i = f'{self.identifier}'
        return f"<{self.__class__.__name__}({self.idname}={i})>"


class Hierarchy(Graphable):
    parents = []
    factors = []
    is_template = True

    def make_specification(self) -> Tuple[Dict[str, Type[Graphable]], Dict[str, str]]:
        """
        Make a dictionary of {name: HierarchyClass} and a similar dictionary of factors
        """
        parents = {p.singular_name if isinstance(p, (type, One2One)) else p.name: p for p in self.parents}
        factors = {f.lower(): f for f in self.factors}
        specification = parents.copy()
        specification.update(factors)
        return specification, factors

    def __init__(self, do_not_create=False, tables=None, **kwargs):
        self.uses_tables = False
        if tables is None:
            for value in kwargs.values():
                if isinstance(value, Unwind):
                    self.uses_tables = True
                elif isinstance(value, Hierarchy):
                    self.uses_tables = value.uses_tables
        else:
            self.uses_tables = True
        self.identifier = kwargs.pop(self.idname, None)
        self.specification, factors = self.make_specification()
        # add any data held in a neo4j unwind table
        for k, v in self.specification.items():
            if k not in kwargs:
                if tables is not None:
                    kwargs[k] = tables.get(k, alias=False)
        self._kwargs = kwargs.copy()
        # Make predecessors a dict of {name: [instances of required Factor/Hierarchy]}
        predecessors = {}
        for name, nodetype in self.specification.items():
            if do_not_create:
                value = kwargs.pop(name, None)
            else:
                value = kwargs.pop(name)
            setattr(self, name, value)
            if isinstance(nodetype, Multiple) and not isinstance(nodetype, One2One):
                if not isinstance(value, (tuple, list)):
                    if isinstance(value, Graphable):
                        if not getattr(value, 'uses_tables', False):
                            raise TypeError(f"{name} expects multiple elements")
            else:
                value = [value]
            if name not in factors:
                predecessors[name] = value
        if len(kwargs):
            raise KeyError(f"{kwargs.keys()} are not relevant to {self.__class__}")
        self.predecessors = predecessors
        if self.identifier_builder is not None:
            if self.identifier is not None:
                raise RuleBreakingException(f"{self} must not take an identifier if it has an identifier_builder")
        if self.idname is not None:
            if not do_not_create and self.identifier is None:
                raise ValueError(f"Cannot assign an id of None to {self}")
            setattr(self, self.idname, self.identifier)
        super(Hierarchy, self).__init__(do_not_create, **predecessors)



from collections import defaultdict
from functools import reduce
from operator import and_
from typing import List

import networkx as nx

from weaveio.basequery.tree import Branch
from weaveio.basequery.actions import DataReference, Alignment
from weaveio.writequery import CypherQuery
from weaveio.writequery.base import Statement, CypherData


def flatten(S):
    if S == []:
        return S
    if isinstance(S[0], list):
        return flatten(S[0]) + flatten(S[1:])
    return S[:1] + flatten(S[1:])


def print_nested_list(nested, tab=0):
    for entry in nested:
        if isinstance(entry, list):
            print_nested_list(['--'] + entry, tab+1)
        else:
            print('    '*tab, entry)


def shared_hierarchy_branch(graph, branch: Branch):
    branches = reduce(and_, [set(p.find_hierarchy_branches()) for p in branch.parents])
    distances = [(b, nx.shortest_path_length(graph, b, branch)) for b in branches]
    try:
        return max(distances, key=lambda x: x[1])[0]
    except ValueError:
        return branch.handler.entry  # no shared parents means going back to the beginning


def parse(graph) -> List:
    aligns = [i for i in graph.nodes if isinstance(i.action, Alignment)][::-1]
    shared_aligns = defaultdict(list)
    for align in aligns:
        shared_aligns[shared_hierarchy_branch(graph, align)].append(align)
    query = []
    todo = list(nx.algorithms.topological_sort(graph))
    while todo:
        node = todo.pop(0)
        if node in shared_aligns:
            align_list = shared_aligns[node][:1]
            query.append(node)
            for align in align_list:
                inputs = align.action.branches
                inputs += (align.action.reference, )
                subqueries = []
                for input_node in inputs:
                    before = list(nx.descendants(graph, node)) + [node]
                    after = list(nx.ancestors(graph, input_node)) + [input_node]
                    newgraph = nx.subgraph_view(graph, lambda n: n in before and n in after)
                    subquery = parse(newgraph)[1:]
                    subqueries.append(subquery)
                done = list(set(flatten(subqueries)))
                for d in done:
                    if d in todo:
                        del todo[todo.index(d)]
                if align.action.continues_on:
                    reference_subquery = subqueries.pop(-1)
                query += subqueries
                if align.action.continues_on:
                    query += reference_subquery
        else:
            query.append(node)
    return query

class OpenSubquery(Statement):
    def __init__(self, input_variables, output_variables, hidden_variables=None):
        super().__init__(input_variables, output_variables, hidden_variables)

    def to_cypher(self):
        inputs = ', '.join(map(str, ['time0'] + self.input_variables))
        return f"CALL {{with {inputs}"


class CloseSubquery(Statement):
    def __init__(self, input_variables, output_variables, hidden_variables=None):
        super().__init__(input_variables, output_variables, hidden_variables)

    def to_cypher(self):
        inputs = ', '.join(map(str, self.output_variables))
        return f"RETURN {inputs}\n}}"


def write_tree(parsed_tree):
    query = CypherQuery.get_context()  # type: CypherQuery
    if isinstance(parsed_tree, list):
        inputs = [i for n in flatten(parsed_tree) for i in n.action.input_variables]
        outputs = [i for n in flatten(parsed_tree) for i in n.action.output_variables]
        subquery_inputs = list({i for i in inputs if getattr(i, 'parent', i) not in outputs and not isinstance(i, CypherData)})

        open = OpenSubquery(subquery_inputs, [])
        query.add_statement(open, safe=False)
        for node in parsed_tree:
            subquery_output = write_tree(node)
        subquery_output = list({v for v in subquery_output if v not in subquery_inputs})
        close = CloseSubquery(subquery_output, subquery_output)
        query.add_statement(close)
        return subquery_output
    else:
        query.add_statement(parsed_tree.action, safe=False)
        return parsed_tree.find_variables()

def branch2query(branch) -> CypherQuery:
    graph = branch.relevant_graph
    subqueries = parse(graph)
    with CypherQuery() as query:
        for node in graph.nodes:
            if isinstance(node.action, DataReference):
                query.data += node.action.input_variables
        for s in subqueries:
            if isinstance(s, list):
                if len(s) == 0:
                    continue
            write_tree(s)
    return query


from weaveio.data import Data as BaseData
from weaveio.opr3.l1files import RawFile, L1SingleFile, L1StackFile, L1SuperStackFile, L1SuperTargetFile
from weaveio.opr3.l2files import L2StackFile, L2SuperStackFile, L2SingleFile, L2SuperTargetFile


class Data(BaseData):
    filetypes = [RawFile, L1SingleFile, L2SingleFile, L1StackFile, L2StackFile, L1SuperStackFile, L2SuperStackFile, L1SuperTargetFile, L2SuperTargetFile]


from pathlib import Path
from typing import Union

from astropy.io import fits
from astropy.table import Table as AstropyTable

from weaveio.config_tables import progtemp_config
from weaveio.file import File, PrimaryHDU, TableHDU, SpectralBlockHDU, SpectralRowableBlock, BinaryHDU
from weaveio.hierarchy import unwind, collect, Multiple, One2One
from weaveio.opr3.hierarchy import Survey, SubProgramme, SurveyCatalogue, \
    WeaveTarget, SurveyTarget, Fibre, FibreTarget, ProgTemp, ArmConfig, ObsTemp, \
    OBSpec, OB, Exposure, Run, Observation, RawSpectrum, L1SingleSpectrum, L1StackSpectrum, L1SuperStackSpectrum, L1SuperTargetSpectrum, CASU
from weaveio.writequery import groupby, CypherData


class HeaderFibinfoFile(File):
    is_template = True

    @classmethod
    def length(cls, path):
        return len(cls.read_fibinfo_dataframe(path))

    @classmethod
    def read_fibinfo_dataframe(cls, path, slc=None):
        hdus = fits.open(path)
        fibinfo_hdu = [i for i in hdus if i.name == 'FIBTABLE'][0]
        fibinfo = AstropyTable(fibinfo_hdu.data).to_pandas()
        fibinfo.columns = [i.lower() for i in fibinfo.columns]
        if 'nspec' in fibinfo.columns:
            fibinfo['spec_index'] = fibinfo['nspec'] - 1
        fibinfo = fibinfo[fibinfo.cname != '']  # exclude parked or disabled fibres
        slc = slice(None) if slc is None else slc
        return fibinfo.iloc[slc]

    @classmethod
    def read_surveyinfo(cls, df_fibinfo):
        df_svryinfo = df_fibinfo[['targsrvy', 'targprog', 'targcat']].drop_duplicates()
        df_svryinfo['progid'] = df_svryinfo['targsrvy'] + df_svryinfo['targprog']
        df_svryinfo['catid'] = df_svryinfo['progid'] + df_svryinfo['targcat']
        return df_svryinfo

    @classmethod
    def read_fibretargets(cls, obspec, df_svryinfo, df_fibinfo):
        srvyinfo = CypherData(df_svryinfo)
        fibinfo = CypherData(df_fibinfo)
        with unwind(srvyinfo) as svryrow:
            with unwind(svryrow['targsrvy']) as surveyname:
                survey = Survey(name=surveyname)
            surveys = collect(survey)
            prog = SubProgramme(name=svryrow['targprog'], progid=svryrow['progid'], surveys=surveys)
            cat = SurveyCatalogue(name=svryrow['targcat'], catid=svryrow['catid'], subprogramme=prog)
        cat_collection = collect(cat)
        cats = groupby(cat_collection, 'name')
        with unwind(fibinfo) as fibrow:
            fibre = Fibre(fibreid=fibrow['fibreid'])  #  must be up here for some reason otherwise, there will be duplicates
            cat = cats[fibrow['targcat']]
            weavetarget = WeaveTarget(cname=fibrow['cname'])
            surveytarget = SurveyTarget(surveycatalogue=cat, weavetarget=weavetarget, tables=fibrow)
            fibtarget = FibreTarget(obspec=obspec, surveytarget=surveytarget, fibre=fibre, tables=fibrow)
        return collect(fibtarget, fibrow)

    @classmethod
    def read_distinct_survey_info(cls, df_svryinfo):
        rows = CypherData(df_svryinfo['targsrvy'].drop_duplicates().values, 'surveynames')
        with unwind(rows) as survey_name:
            survey = Survey(name=survey_name)
        survey_list= collect(survey)
        survey_dict = groupby(survey_list, 'name')

        # each row is (subprogramme, [survey_name, ...])
        s = df_svryinfo.groupby('progid')[['targsrvy', 'targprog']].aggregate(lambda x: x.values.tolist())
        rows = CypherData(s, 'targprog_rows')
        with unwind(rows) as row:
            with unwind(row['targsrvy']) as survey_name:
                survey = survey_dict[survey_name]
            surveys = collect(survey)
            subprogramme = SubProgramme(surveys=surveys, name=row['targprog'][0], progid=row['progid'])
        subprogramme_list = collect(subprogramme)
        subprogramme_dict = groupby(subprogramme_list, 'progid')

        # catalogue has 1 programme
        # programme can have many catalogues
        rows = CypherData(df_svryinfo[['targcat', 'progid', 'catid']].drop_duplicates(), 'targcat_rows')
        with unwind(rows) as row:
            catalogue = SurveyCatalogue(subprogramme=subprogramme_dict[row['progid']], name=row['targcat'], catid=row['catid'])
        catalogue_list = collect(catalogue)
        return survey_list, subprogramme_list, catalogue_list

    @classmethod
    def read_hierarchy(cls, header, df_svryinfo):
        surveys, subprogrammes, catalogues = cls.read_distinct_survey_info(df_svryinfo)
        runid = int(header['RUN'])
        camera = str(header['CAMERA'].lower()[len('WEAVE'):])
        expmjd = float(header['MJD-OBS'])
        res = str(header['VPH']).rstrip('123')
        obstart = float(header['OBSTART'])
        obtitle = str(header['OBTITLE'])
        xml = str(header['CAT-NAME'])
        obid = float(header['OBID'])

        progtemp = ProgTemp.from_progtemp_code(header['PROGTEMP'])
        vph = int(progtemp_config[(progtemp_config['mode'] == progtemp.instrumentconfiguration.mode)
                                  & (progtemp_config['resolution'] == res)][f'{camera}_vph'].iloc[0])
        arm = ArmConfig(vph=vph, resolution=res, camera=camera)  # must instantiate even if not used
        obstemp = ObsTemp.from_header(header)
        obspec = OBSpec(xml=xml, obtitle=obtitle, obstemp=obstemp, progtemp=progtemp,
                        surveycatalogues=catalogues, subprogrammes=subprogrammes, surveys=surveys)
        ob = OB(obid=obid, obstartmjd=obstart, obspec=obspec)
        exposure = Exposure(expmjd=expmjd, ob=ob)
        run = Run(runid=runid, armconfig=arm, exposure=exposure)
        observation = Observation.from_header(run, header)
        return {'run': run, 'ob': ob, 'obspec': obspec, 'armconfig': arm, 'observation': observation}

    @classmethod
    def read_schema(cls, path: Path, slc: slice = None):
        header = cls.read_header(path)
        fibinfo = cls.read_fibinfo_dataframe(path, slc)
        srvyinfo = cls.read_surveyinfo(fibinfo)
        hiers = cls.read_hierarchy(header, srvyinfo)
        fibretarget_collection, fibrows = cls.read_fibretargets(hiers['obspec'], srvyinfo, fibinfo)
        return hiers, header, fibinfo, fibretarget_collection, fibrows

    @classmethod
    def read_header(cls, path: Path):
        return fits.open(path)[0].header

    @classmethod
    def read_fibtable(cls, path: Path):
        return AstropyTable(fits.open(path)[cls.fibinfo_i].data)

    @classmethod
    def read(cls, directory: Path, fname: Path, slc: slice = None) -> 'File':
        raise NotImplementedError


class RawFile(HeaderFibinfoFile):
    match_pattern = 'r*.fit'
    hdus = {'primary': PrimaryHDU, 'counts1': SpectralBlockHDU, 'counts2': SpectralBlockHDU, 'fibtable': TableHDU, 'guidinfo': TableHDU, 'metinfo': TableHDU}
    parents = [CASU]
    produces = [One2One(Observation), One2One(RawSpectrum)]

    @classmethod
    def fname_from_runid(cls, runid):
        return f'r{runid:07.0f}.fit'

    @classmethod
    def read(cls, directory: Union[Path, str], fname: Union[Path, str], slc: slice = None):
        path = Path(directory) / Path(fname)
        hiers, header, fibinfo, fibretarget_collection, fibrow_collection = cls.read_schema(path, slc)
        observation = hiers['observation']
        hdus, file, _ = cls.read_hdus(directory, fname, casu=observation.casu)
        raw = RawSpectrum(sourcefile=str(fname), casu=observation.casu, observation=observation, nrow=-1)
        raw.attach_products(file, **hdus)
        observation.attach_products(file, **hdus)
        return file


class L1File(HeaderFibinfoFile):
    is_template = True
    hdus = {'primary': PrimaryHDU, 'flux': SpectralRowableBlock, 'ivar': SpectralRowableBlock,
            'flux_noss': SpectralRowableBlock, 'ivar_noss': SpectralRowableBlock,
            'sensfunc': BinaryHDU, 'fibtable': TableHDU}


class L1SingleFile(L1File):
    match_pattern = 'single_*.fit'
    parents = [One2One(RawFile), CASU]
    produces = [L1SingleSpectrum]
    version_on = ['rawfile']

    @classmethod
    def fname_from_runid(cls, runid):
        return f'single_{runid:07.0f}.fit'

    @classmethod
    def read(cls, directory: Union[Path, str], fname: Union[Path, str], slc: slice = None):
        fname = Path(fname)
        directory = Path(directory)
        path = directory / fname
        hiers, header, fibinfo, fibretarget_collection, fibrow_collection = cls.read_schema(path, slc)
        observation = hiers['observation']
        casu = observation.casu
        inferred_raw_fname = str(fname.with_name(fname.name.replace('single_', 'r')))
        raw = RawSpectrum(sourcefile=inferred_raw_fname, casu=casu, observation=observation, nrow=-1)
        rawfile = RawFile(inferred_raw_fname, casu=casu)  # merge this one instead of finding, then we can start from single or raw files
        hdus, file, _ = cls.read_hdus(directory, fname, rawfile=rawfile, casu=casu)
        with unwind(fibretarget_collection, fibrow_collection) as (fibretarget, fibrow):
            single_spectrum = L1SingleSpectrum(sourcefile=str(fname), nrow=fibrow['nspec'],
                                               rawspectrum=raw, fibretarget=fibretarget,
                                               casu=casu, tables=fibrow)
            single_spectrum.attach_products(file, index=fibrow['spec_index'], **hdus)
        single_spectra = collect(single_spectrum)  # must collect at the end
        return file


class L1StackedBaseFile(L1File):
    is_template = True


class L1StackFile(L1StackedBaseFile):
    match_pattern = 'stacked_*.fit'
    produces = [L1StackSpectrum]
    parents = [Multiple(L1SingleFile), OB, ArmConfig, CASU]

    @classmethod
    def fname_from_runid(cls, runid):
        return f'stacked_{runid:07.0f}.fit'

    @classmethod
    def parent_runids(cls, path):
        header = cls.read_header(path)
        return [int(v) for k, v in header.items() if k.startswith('RUNS0')]

    @classmethod
    def get_single_files(cls, directory: Path, fname: Path):
        l1singlefiles = []
        runids = cls.parent_runids(directory / fname)
        for runid in runids:
            single_fname = L1SingleFile.fname_from_runid(runid)
            subdir = fname.parents[0]
            l1singlefiles.append(L1SingleFile.find(fname=str(subdir / single_fname)))
        return l1singlefiles

    @classmethod
    def read(cls, directory: Union[Path, str], fname: Union[Path, str], slc: slice = None):
        """
        L1Stack inherits everything from the lowest numbered single/raw files so we are missing data,
        therefore we require that all the referenced Runs are present before loading in
        """
        fname = Path(fname)
        directory = Path(directory)
        path = directory / fname
        hiers, header, fibinfo, fibretarget_collection, fibrow_collection = cls.read_schema(path, slc)

        observation = hiers['observation']
        ob = hiers['ob']
        armconfig = hiers['armconfig']
        casu = observation.casu
        singlefiles = cls.get_single_files(directory, fname)
        hdus, file, _ = cls.read_hdus(directory, fname, l1singlefiles=singlefiles, ob=ob,
                                   armconfig=armconfig, casu=casu)
        with unwind(fibretarget_collection, fibrow_collection) as (fibretarget, fibrow):
            single_spectra = []
            for singlefile in singlefiles:
                single_spectrum = L1SingleSpectrum.find(sourcefile=str(singlefile.fname), nrow=fibrow['nspec'])
                single_spectra.append(single_spectrum)
            stack_spectrum = L1StackSpectrum(sourcefile=str(fname), nrow=fibrow['nspec'],
                                             l1singlespectra=single_spectra, ob=ob,
                                             armconfig=armconfig, fibretarget=fibretarget,
                                             casu=casu, tables=fibrow)
            stack_spectrum.attach_products(file, index=fibrow['spec_index'], **hdus)
        stack_spectra = collect(stack_spectrum)  # must collect at the end
        return file


class L1SuperStackFile(L1StackedBaseFile):
    match_pattern = 'superstacked_*.fit'
    produces = [L1SuperStackSpectrum]
    parents = [Multiple(L1SingleFile), OBSpec, ArmConfig, CASU]

    @classmethod
    def fname_from_runid(cls, runid):
        return f'superstacked_{runid:07.0f}.fit'

    @classmethod
    def match(cls, directory):
        return directory.glob()

    @classmethod
    def read(cls, directory: Union[Path, str], fname: Union[Path, str], slc: slice = None):
        raise NotImplementedError


class L1SuperTargetFile(L1StackedBaseFile):
    match_pattern = 'WVE_*.fit'
    parents = [WeaveTarget, Multiple(L1SingleFile, 2), CASU]
    produces = [L1SuperTargetSpectrum]
    recommended_batchsize = None

    @classmethod
    def read(cls, directory: Union[Path, str], fname: Union[Path, str], slc: slice = None):
        raise NotImplementedError





from pathlib import Path

from .__version__ import __version__

import logging

try:
    from boxing import boxing
    from colored import fore, style
    from pkg_info import get_pkg_info
    from semver import compare
    import urllib
    import json
    from datetime import datetime
    from typing import Callable


    def parse_changes(description):
        log = description.split('machine-readable-change-log\n###########################\n')[-1]
        return log


    def get_other_descriptions(name, version):
        return json.loads(urllib.request.urlopen(f'https://pypi.org/pypi/{name}/{version}/json', timeout=3).read())['info']['description']


    class UpdateNotify(object):
        def __init__(self, name: str, version: str):
            self.name: str = name
            self.version: str = version
            self.last_checked_path = Path(__file__).parent / '.last-checked'
            self.time_fmt = '%Y-%m-%dT%H:%M:%S'
            self._pkg = None

        @property
        def pkg(self):
            if self._pkg is None:
                self._pkg = get_pkg_info(self.name)
            return self._pkg

        def last_checked(self):
            try:
                with open(str(self.last_checked_path), 'r') as f:
                    return datetime.strptime(f.read().strip(), self.time_fmt)
            except FileNotFoundError:
                return None

        def too_soon(self):
            lc = self.last_checked()
            if lc is None:
                return False
            seconds = (datetime.now() - lc).total_seconds()
            return (seconds / 60 / 60) < 6

        def update_last_checked(self):
            with open(str(self.last_checked_path), 'w') as f:
                f.write(datetime.now().strftime(self.time_fmt))

        def is_latest_version(self) -> bool:
            self.latest = self.pkg.version
            return True if compare(self.version, self.latest) >= 0 else False

        def render_changes(self):
            releases = [(k, datetime.strptime(v[0]['upload_time'], self.time_fmt)) for k, v in self.pkg.raw_data['releases'].items()]
            releases.sort(key=lambda x: x[1])
            release_names, _ = zip(*releases)
            changes = [parse_changes(get_other_descriptions(self.pkg.name, v)) for v in release_names]
            newer_changes = changes[release_names.index(self.pkg.version):]
            return '\n'.join(newer_changes), len(newer_changes)

        def notify(self) -> None:
            if self.too_soon():
                logging.info('Skipping version checking since its has not been 6 hours since the last check.')
                return
            self.update_last_checked()
            if self.is_latest_version():
                return
            action, arg = print, self.default_message()
            action(arg) if arg else action()

        def default_message(self) -> str:
            changes, nchanges = self.render_changes()
            version = fore.GREY_53 + self.version + style.RESET
            latest = fore.LIGHT_GREEN + self.latest + style.RESET
            command = fore.LIGHT_BLUE + 'pip install -U ' + self.name + style.RESET
            nchanges = fore.LIGHT_GREEN + str(nchanges) + style.RESET
            return boxing(f'Update available {version} → {latest} ({nchanges} new changes)\n' +
                          f'Run {command} to update\n'
                          f'{changes}')

    UpdateNotify('weaveio', __version__).notify()

except ImportError:
    from warnings import warn
    warn('Please run `pip install boxing colored pkg_info semver` to alert you to updated versions of the weaveio library')
except Exception as e:
    logging.exception('There was a problem in alerting you to updated versions of the weaveio library...', exc_info=True)



from collections import defaultdict
from pathlib import Path
from typing import Union, List, Dict, Type, Set, Tuple

from astropy.io import fits
from astropy.io.fits.hdu.base import _BaseHDU
from astropy.table import Table

from weaveio.file import File, PrimaryHDU, TableHDU
from weaveio.graph import Graph
from weaveio.hierarchy import Multiple, unwind, collect, Hierarchy
from weaveio.opr3.hierarchy import APS, L1SpectrumRow, FibreTarget, OB, OBSpec, L2Stack, L2SuperStack, L2SuperTarget, L2Single, ClassificationTable, GalaxyTable, Exposure, WeaveTarget, L2, ClassificationSpectrum, GalaxySpectrum
from weaveio.opr3.l1files import L1File, L1SuperStackFile, L1StackFile, L1SingleFile, L1SuperTargetFile
from weaveio.writequery import CypherData, groupby


class MissingDataError(Exception):
    pass


def filter_products_from_table(table: Table, maxlength: int) -> Table:
    columns = []
    for i in table.colnames:
        value = table[i]
        if len(value.shape) == 2:
            if value.shape[1] > maxlength:
                continue
        columns.append(i)
    return table[columns]


class L2File(File):
    is_template = True
    match_pattern = '*aps.fits'
    produces = [L2]
    corresponding_hdus = {'class_table': ClassificationTable, 'galaxy_table': GalaxyTable,
                          'class_spectra': ClassificationSpectrum, 'galaxy_spectra': GalaxySpectrum}
    parents = [Multiple(L1File, 2, 3), APS]
    hdus = {'primary': PrimaryHDU, 'fibtable': TableHDU,
            'class_spectra': TableHDU,
            'stellar_spectra_ferre': TableHDU, 'stellar_spectra_rvs': TableHDU,
            'galaxy_spectra': TableHDU,
            'class_table': TableHDU,
            'stellar_table': TableHDU,
            'stellar_table_rvs': TableHDU,
            'galaxy_table': TableHDU}

    @classmethod
    def length(cls, path):
        hdus = fits.open(path)
        names = [i.name for i in hdus]
        return len(hdus[names.index('CLASS_SPECTRA')].data)

    @classmethod
    def decide_filetype(cls, l1filetypes: List[Type[File]]) -> Type[File]:
        l1precedence = [L1SingleFile, L1StackFile, L1SuperStackFile, L1SuperTargetFile]
        l2precedence = [L2SingleFile, L2StackFile, L2SuperStackFile, L2SuperTargetFile]
        highest = max(l1precedence.index(l1filetype) for l1filetype in l1filetypes)
        return l2precedence[highest]

    @classmethod
    def match_file(cls, directory: Union[Path, str], fname: Union[Path, str], graph: Graph):
        """
        L2 files can be formed from any combination of L1 files and so the shared hierarchy level can be
        either exposure, OB, OBSpec, or WeaveTarget.
        L2 files are distinguished by the shared hierarchy level of their formative L1 files.
        Therefore, we assign an L2 file to the highest hierarchy level.
        e.g.
        L1Single+L1Single -> L2Single
        L1Stack+L1Single -> L2Stack
        L1SuperStack+L1Stack -> L2SuperStack
        """
        fname = Path(fname)
        directory = Path(directory)
        path = directory / fname
        if not super().match_file(directory, fname, graph):
            return False
        header = cls.read_header(path)
        ftypes, _ = zip(*cls.parse_fname(header, fname, instantiate=False))
        return cls.decide_filetype(ftypes) is cls

    @classmethod
    def parse_fname(cls, header, fname, instantiate=True) -> List[L1File]:
        """
        Return the L1File type and the expected filename that formed this l2 file
        """
        ftype_dict = {
            'single': L1SingleFile,
            'stacked': L1StackFile, 'stack': L1StackFile,
            'superstack': L1SuperStackFile, 'superstacked': L1SuperStackFile
        }
        split = fname.name.replace('.aps.fits', '').replace('.aps.fit', '').split('_')
        runids = []
        ftypes = []
        for i in split:
            try:
                runids.append(int(i))
            except ValueError:
                ftypes.append(str(i))
        if len(ftypes) == 1:
            ftypes = [ftypes[0]] * len(runids)  # they all have the same type if there is only one mentioned
        assert len(ftypes) == len(runids), "error parsing runids/types from fname"
        assert all(int(i) in runids for i in header['RUN'].split('+')), "fname runids and header runids do not match"
        files = []
        for ftype, runid in zip(ftypes, runids):
            ftype_cls = ftype_dict[ftype]
            fname = ftype_cls.fname_from_runid(runid)
            if instantiate:
                files.append(ftype_cls.find(fname=fname))
            else:
                files.append((ftype_cls, fname))
        return files

    @classmethod
    def find_shared_hierarchy(cls, path: Path) -> Dict:
        raise NotImplementedError

    @classmethod
    def read_header(cls, path):
        return fits.open(path)[0].header

    @classmethod
    def get_all_nspecs(cls, astropy_hdus, slc) -> List[int]:
        nspecs = set()
        for hdu in astropy_hdus:
            try:
                nspecs.update(hdu.data['nspec'].tolist())
            except (TypeError, ):  # dont do the primary header
                pass
        return list(nspecs)[slc]

    @classmethod
    def read_hdus(cls, directory: Union[Path, str], fname: Union[Path, str], l1files: List[L1File],
                  **hierarchies: Union[Hierarchy, List[Hierarchy]]) -> Tuple[Dict[str, 'HDU'], 'File', List[_BaseHDU]]:
        fdict = {p.plural_name: [] for p in cls.parents if isinstance(p, Multiple) and issubclass(p.node, L1File)} # parse the 1lfile types separately
        for f in l1files:
            fdict[f.plural_name].append(f)
        hierarchies.update(fdict)
        return super().read_hdus(directory, fname, **hierarchies)

    @classmethod
    def produce_l2(cls, sourcefile, nrow, l1spectrumrows, aps, **hierarchies):
        assert len(cls.produces) == 1
        sdict = {p.plural_name: [] for p in cls.produces[0].parents if isinstance(p, Multiple) and issubclass(p.node, L1SpectrumRow)}  # parse the l1spectrum types separately
        for f in l1spectrumrows:
            sdict[f.plural_name].append(f)
        hierarchies.update(sdict)
        l2 = cls.produces[0](sourcefile=sourcefile.fname, nrow=nrow, aps=aps, **hierarchies)
        l2.attach_products(sourcefile)
        return l2

    @classmethod
    def read(cls, directory: Union[Path, str], fname: Union[Path, str], slc: slice = None):
        fname = Path(fname)
        directory = Path(directory)
        path = directory / fname
        header = cls.read_header(path)
        l1files = cls.parse_fname(header, fname)
        aps = APS(apsvers=header['APSVERS'])
        hierarchies = cls.find_shared_hierarchy(path)
        hdu_nodes, file, astropy_hdus = cls.read_hdus(directory, fname, l1files=l1files, aps=aps, **hierarchies)
        nspecs = cls.get_all_nspecs(astropy_hdus, slc)
        with unwind(CypherData(nspecs, 'nspecs')) as nspec:
            l1spectra = []
            for l1file in l1files:
                spectrum = l1file.produces[0].find(sourcefile=l1file.fname, nrow=nspec)
                l1spectra.append(spectrum)
            if not issubclass(cls, L2SuperTargetFile):
                fibretarget = FibreTarget.find(anonymous_children=[l1spectra[0]])
                hierarchies['fibretarget'] = fibretarget
            l2 = cls.produce_l2(sourcefile=file, nrow=nspec, l1spectrumrows=l1spectra, aps=aps, **hierarchies)
        l2s = collect(l2)
        l2_dict = groupby(l2s, 'nrow')
        for name in cls.corresponding_hdus.keys():
            cls.make_data_rows(name, nspecs, file, l2_dict, astropy_hdus, hdu_nodes)
        return file

    @classmethod
    def read_one_hdu_l2data(cls, hdus, hduname, nspecs):
        names = [i.name.lower().strip() for i in hdus]
        table = Table(hdus[names.index(hduname)].data)
        if len(table.colnames):
            table.rename_columns(table.colnames, [i.lower() for i in table.colnames])
            table['spec_index'] = range(len(table))
        table = filter_products_from_table(table, 10)  # removes huge arrays that are best kept in binary files
        if len(table):
            table = table[[i in nspecs for i in table['nspec']]]
        data = CypherData(table, hduname)
        return data

    @classmethod
    def make_data_rows(cls, hduname, nspecs: List[int], file, l2_by_nrow, astropyhdulist, hdus):
        """
        For each L2 row, there are 2-3 L1 rows
        Since we know the L1 files, we can identify those rows with nspec since they are the same
        """
        row_type = cls.corresponding_hdus[hduname]
        table = cls.read_one_hdu_l2data(astropyhdulist, hduname, nspecs)
        if len(table.data):  # to avoid cutting off the query
            with unwind(table) as row:
                nspec = row['nspec']
                l2row = row_type(sourcefile=file.fname, nrow=nspec, hduname=hduname, l2=l2_by_nrow[nspec], tables=row)
                if len(row_type.products):
                    l2row.attach_products(file, index=row['spec_index'], **hdus)
            l2rows = collect(l2row)
            return l2rows
        return None


class L2SingleFile(L2File):
    produces = [L2Single]
    parents = [Multiple(L1SingleFile, 2, 3), Exposure, APS]

    @classmethod
    def find_shared_hierarchy(cls, path) -> Dict:
        header = cls.read_header(path)
        return {'exposure': Exposure.find(obid=header['MJD-OBS'])}


class L2StackFile(L2File):
    produces = [L2Stack]
    parents = [Multiple(L1SingleFile, 0, 3), Multiple(L1StackFile, 1, 3), OB, APS]

    @classmethod
    def find_shared_hierarchy(cls, path) -> Dict:
        header = cls.read_header(path)
        return {'ob': OB.find(obid=header['OBID'])}


class L2SuperStackFile(L2File):
    produces = [L2SuperStack]
    parents = [Multiple(L1SingleFile, 0, 3), Multiple(L1StackFile, 0, 3), Multiple(L1SuperStackFile, 0, 3), OBSpec, APS]

    @classmethod
    def find_shared_hierarchy(cls, path) -> Dict:
        header = cls.read_header(path)
        return {'obspec': OBSpec.find(xml=str(header['cat-name']))}


class L2SuperTargetFile(L2File):
    match_pattern = 'WVE_*aps.fits'
    produces = [L2SuperTarget]
    parents = [Multiple(L1SuperTargetFile, 2, 3), WeaveTarget, APS]

    @classmethod
    def parse_fname(cls, header, fname, instantiate=True) -> List[L1File]:
        raise NotImplementedError

    @classmethod
    def find_shared_hierarchy(cls, path: Path) -> Dict:
        hdus = fits.open(path)
        names = [i.name for i in hdus]
        cname = hdus[names.index('CLASS_TABLE')].data['CNAME'][0]
        return {'weavetarget': WeaveTarget.find(cname=cname)}
